//File:         EventList.java
//Project:      ProbeSim
//Version:      0.2
//Designer:     C. Edward Chow
//Revision:
//              1/14/2000  C. Edward Chow
//Bug Report:   send to chow@cs.uccs.edu
//Description:  Simulator for Internet Network Probing Research
//              Efficent Discrete Event Queue Class
//

package NetLayout;

import java.util.*;
import java.io.*;

class EventListNode {
  double idx;
  EventListNode skip, next, prev;
  EventRecord er;
  EventListNode() {
  }
  EventListNode(EventRecord er) {
    this.er = er;
    skip = null;
    next = null;
    prev = null;
  }
}

public class EventList {
  double currentTime;
  EventListNode eventList;
  boolean saveEvents;
  EventListNode dummyNode;
  EventRecord dummyEvent;
  public EventList() {
    dummyEvent = new EventRecord();
    dummyEvent.time = Float.POSITIVE_INFINITY;
    dummyEvent.next = -1;
    dummyEvent.src = -1;
    dummyEvent.dst = -1;
    dummyEvent.msg = null;

    dummyNode = new EventListNode();
    dummyNode.idx = Float.POSITIVE_INFINITY;
    dummyNode.next = null;
    dummyNode.prev = null;
    dummyNode.skip = null;
    dummyNode.er = dummyEvent;

    eventList = dummyNode;
    currentTime = 0.0;
    saveEvents = false;
  }
  public void insertEvent(EventRecord x, EventListNode current) {
    EventListNode temp = new EventListNode(x);

    temp.prev = current.prev;
    temp.next = current;

    /* set indexing fields */
    temp.skip = current;
    temp.idx = current.er.time;

    /* fix previous node to point to new node,
     unless current is present head of list */
    if (current.prev != null)  current.prev.next = temp;

    /* fix next node to point to new node */
    current.prev = temp;

    /* inserted at head of list, replace eventList */
    if (current == eventList)
    eventList = temp;

    /* inserted at tail of list, add shortcut to prevent
     O( N^2 ) performance for (nearly) ordered input by
     always having first node's index point to last node */

    if (current == dummyNode && eventList != temp) {
      /*NOTE: above test also prevents loop when only one node on list */
      eventList.skip = temp;
      eventList.idx = temp.er.time;
    }
  }

  public void queueEvent(EventRecord er) {
    EventListNode current;

    if (er.time < currentTime) {
      System.out.println("%%NETSIM - Event with illegal time stamp"+er.time+" received by Queue_Event.\n");
      return;
    }
    current = eventList;
    while (current.next != null) {   /*not at end of list... */
      if (er.time >= current.idx)   /* use skip pointer to bypass N nodes */
        current = current.skip;
      else {  /* check the node's event time value */
        if (er.time < current.er.time) {   /* not found yet... */
	        insertEvent(er, current);
	        return;
        }
        current = current.next;
      }
    }
    insertEvent(er, current);
  }
  public EventRecord nextEvent(){
    EventRecord result;

    do {
      currentTime = eventList.er.time;
      result = eventList.er;
      eventList = eventList.next;
      if (!saveEvents) {
         // Free(eventList.prev);  no need to free garbage collection handle that
        eventList.prev = null;
      }
    } while (currentTime < 0.0);

    /*i.e., non-cancelled event,
      dummy event always triggers this condition*/

    if (eventList == null) {
     /* was at end of list, just read dummy event,  return NIL event pointer */
      result = null;
      eventList = dummyNode;   /*reset eventList to point to 'empty' list*/
    }
    return result;
  }
  public void cancelEvents(int src){
    EventListNode current;

    current = eventList;
    while (current != dummyNode) {   /*at end of list... */
      if (current.er.src == src)   /* cancel this event by negating time */
        current.er.time = -current.er.time;
      current = current.next;
    }
  }
  double sim_time(){
    return currentTime;
  }
  public void setSaveMode(){
    saveEvents = true;
  }
  public void setEventTracking(Hashtable options){
  }
}

